"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.WorkerPool = void 0;
const async_hooks_1 = require("async_hooks");
const events_1 = require("events");
const path_1 = __importDefault(require("path"));
const worker_threads_1 = require("worker_threads");
const backtrace_service_1 = require("backtrace-service");
const log = backtrace_service_1.Logger.getLogger();
const kWorkerFreedEvent = Symbol('kWorkerFreedEvent');
class WorkerPoolTaskInfo extends async_hooks_1.AsyncResource {
    constructor(callback) {
        super('WorkerPoolTaskInfo');
        this.callback = callback;
    }
    done(err, result) {
        this.runInAsyncScope(this.callback, null, err, result);
        this.emitDestroy(); // `TaskInfo`s are used only once.
    }
}
// T is a request type, IDistanceRequest | IDistanceSummaryRequest
// R is a result type, IDistanceTaskResult | ISummaryRow
class WorkerPool extends events_1.EventEmitter {
    constructor(numThreads) {
        super();
        this.workers = [];
        this.freeWorkers = [];
        for (let i = 0; i < numThreads; i++)
            this.addNewWorker();
        log.info(`Worker pool initialized with ${numThreads} workers.`);
    }
    addNewWorker() {
        const workerRaw = new worker_threads_1.Worker(path_1.default.resolve(__dirname, 'taskProcessor.js')); // intentional "any" temp to build ITaskWorker
        workerRaw.kTaskInfo = null;
        const worker = workerRaw;
        worker.on('message', (result) => {
            var _a;
            // On success: Call the callback with result, remove `TaskInfo`, mark as free.
            (_a = worker.kTaskInfo) === null || _a === void 0 ? void 0 : _a.done(null, result);
            worker.kTaskInfo = null;
            this.freeWorkers.push(worker);
            this.emit(kWorkerFreedEvent);
        });
        worker.on('error', (err) => {
            var _a;
            // On error: Call the callback with error.
            if (worker.kTaskInfo)
                (_a = worker.kTaskInfo) === null || _a === void 0 ? void 0 : _a.done(err, null);
            else
                this.emit('error', err);
            // Remove the worker and start a new Worker to replace it.
            this.workers.splice(this.workers.indexOf(worker), 1);
            this.addNewWorker();
        });
        this.workers.push(worker);
        this.freeWorkers.push(worker);
        this.emit(kWorkerFreedEvent);
    }
    runTask(task, callback) {
        if (this.freeWorkers.length === 0) {
            // No free threads, wait until a worker thread becomes free.
            this.once(kWorkerFreedEvent, () => this.runTask(task, callback));
            return;
        }
        const worker = this.freeWorkers.pop();
        if (!worker) {
            throw new Error('no worker');
        }
        worker.kTaskInfo = new WorkerPoolTaskInfo(callback);
        worker === null || worker === void 0 ? void 0 : worker.postMessage(task);
    }
    close() {
        for (const worker of this.workers)
            worker.terminate();
    }
}
exports.WorkerPool = WorkerPool;
//# sourceMappingURL=workerPool.js.map